////////////////////////////////////
////////////////////////////////////
// Class:       ArgoneutCCInclusiveFilter
// Plugin Type: filter (art v2_10_03)
// File:        ArgoneutCCInclusiveFilter_module.cc
//
// Generated at Mon Apr 30 17:01:51 2018 by Tingjun Yang using cetskelgen
// from cetlib version v3_02_00.
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDFilter.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Run.h"
#include "art/Framework/Principal/SubRun.h"
#include "canvas/Utilities/InputTag.h"
#include "canvas/Persistency/Common/FindManyP.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "lardataobj/RecoBase/Track.h"
#include "lardataobj/RecoBase/Cluster.h"
#include "lardataobj/RecoBase/Vertex.h"
#include "lardataobj/RecoBase/Hit.h"
#include "T962_Objects/MINOS.h"

#include "larcore/Geometry/Geometry.h"
#include "lardata/DetectorInfoServices/DetectorPropertiesService.h"

#include <memory>

// TODO: turn these into a parameter set eventually
const double kFidVolStartX = 3.0;
const double kFidVolEndX = 44.0;
const double kFidVolStartY = -16.0;
const double kFidVolEndY = 16.0;
const double kFidVolStartZ = 6.0;
const double kFidVolEndZ = 86.0;
const double kRSphere = 16.0;
const double kRDiffCut = 12.;
const double kThetaDiffCut = 0.17; 

namespace t962 {
  class ArgoneutCCInclusiveFilter;
}


class t962::ArgoneutCCInclusiveFilter : public art::EDFilter {
public:
  explicit ArgoneutCCInclusiveFilter(fhicl::ParameterSet const & p);
  // The compiler-generated destructor is fine for non-base
  // classes without bare pointers or other resource use.

  // Plugins should not be copied or assigned.
  ArgoneutCCInclusiveFilter(ArgoneutCCInclusiveFilter const &) = delete;
  ArgoneutCCInclusiveFilter(ArgoneutCCInclusiveFilter &&) = delete;
  ArgoneutCCInclusiveFilter & operator = (ArgoneutCCInclusiveFilter const &) = delete;
  ArgoneutCCInclusiveFilter & operator = (ArgoneutCCInclusiveFilter &&) = delete;

  // Required functions.
  bool filter(art::Event & e) override;

  // Selected optional functions.
  void beginJob() override;
  void endJob() override;

private:

  double rdiff_minos;
  double xdiff_minos;
  double ydiff_minos;
  double thetadiff_minos;
  int bestmatch;
  int bestmatchminos;

  double trkvtxx_bestmatch;
  double trkvtxy_bestmatch;
  double trkvtxz_bestmatch;

  double trkendx_bestmatch;
  double trkendy_bestmatch;
  double trkendz_bestmatch;

  double largest_open_angle;
  double trkendx_open_angle;
  double trkendy_open_angle;
  double trkendz_open_angle;

  int vertex_index;

  double vtxx;
  double vtxy;
  double vtxz;

  double nearestz[3];
 
  std::string fMINOSModuleLabel;
 
  // Declare member data here.
  void MatchMINOS(std::vector<art::Ptr<recob::Track>>& tracklist,
                  art::FindManyP<t962::MINOS>& fmmatch,
                  bool isdata);
  
  void VertexSelect(std::vector<art::Ptr<recob::Vertex>>& vertexlist,
		    std::vector<art::Ptr<recob::Track>>& tracklist,
		    bool isdata);
  
  void CrossingTrack(std::vector<art::Ptr<recob::Cluster>>& clusterlist,
		     art::FindManyP<recob::Hit>& fmcls,
		     bool isdata);
  
  void BrokenTrack(std::vector<art::Ptr<recob::Track>>& tracklist,
		   bool isdata);

  bool SimpleMatch(std::vector<art::Ptr<t962::MINOS>>& minoslist,
		   bool isdata);
		    
};

// ------------------------------------------------- 

t962::ArgoneutCCInclusiveFilter::ArgoneutCCInclusiveFilter(fhicl::ParameterSet const & pset) :
  EDFilter(pset),
  fMINOSModuleLabel         (pset.get< std::string >("MINOSModuleLabel")        )
// Initialize member data here.
{
  // Call appropriate produces<>() functions here.
}

// ------------------------------------------------- 

bool t962::ArgoneutCCInclusiveFilter::filter(art::Event & evt)
{

  double isdata = 0;
  if (evt.isRealData()){
    isdata = 1;
  }

  // Reconstructed tracks
  art::Handle< std::vector<recob::Track> > trackListHandle;
  std::vector<art::Ptr<recob::Track> > tracklist;
  if (evt.getByLabel("pmtrack",trackListHandle))
    art::fill_ptr_vector(tracklist, trackListHandle);

  // MINOS matched tracks
  art::FindManyP<t962::MINOS> fmmatch(trackListHandle, evt, "matchtracks");
  if (!fmmatch.isValid()) {
    throw cet::exception("ArgoneutCCInclusiveFilter") << "Cannot find MINOS matched tracks";
  }

  art::Handle< std::vector<t962::MINOS> > minosListHandle;
  std::vector<art::Ptr<t962::MINOS> > minoslist;
  if (evt.getByLabel(fMINOSModuleLabel,minosListHandle))
    art::fill_ptr_vector(minoslist, minosListHandle);

  art::Handle< std::vector<recob::Cluster> > clusterListHandle;
  std::vector<art::Ptr<recob::Cluster> > clusterlist;
  if (evt.getByLabel("trajcluster",clusterListHandle))
    art::fill_ptr_vector(clusterlist, clusterListHandle);

  // hits associated with clusters
  art::FindManyP<recob::Hit> fmcls(clusterListHandle, evt, "trajcluster");
  if (!fmcls.isValid()) {
    throw cet::exception("ArgoneutCCInclusiveFilter") << "Cannot find hits associated with clusters";
  }

  art::Handle< std::vector<recob::Vertex> > vertexListHandle;
  std::vector<art::Ptr<recob::Vertex> > vertexlist;
  if (evt.getByLabel("pmtrack",vertexListHandle))
    art::fill_ptr_vector(vertexlist, vertexListHandle);

  bestmatch = -1;
  bestmatchminos = -1;
  vertex_index = -1;

  trkvtxx_bestmatch = 99999; // vertex of best matched track
  trkvtxy_bestmatch = 99999;
  trkvtxz_bestmatch = 99999;

  trkendx_bestmatch = 99999; // vertex of best matched track
  trkendy_bestmatch = 99999;
  trkendz_bestmatch = 99999;

  largest_open_angle = 0;

  vtxx = 99999;
  vtxy = 99999;
  vtxz = 99999;

  rdiff_minos = 99999;
  xdiff_minos = 99999;
  ydiff_minos = 99999;
  thetadiff_minos = 99999;

  for (int i = 0; i < 3; ++i) nearestz[i] = 99999;

  bool minosMatched = false;
  bool hasVertex = false;
  bool inFidVol = false;
  bool noCrossingTrk = false; 
  bool noBrokenTrk = false;

  MatchMINOS(tracklist, fmmatch, isdata);
  // MINOS match?
  if (rdiff_minos < kRDiffCut && thetadiff_minos < kThetaDiffCut) minosMatched = true;

  VertexSelect(vertexlist, tracklist, isdata);
  // is best reco vertex close enough to the matched track vertex?
  double xdist = trkvtxx_bestmatch - vtxx;
  double ydist = trkvtxy_bestmatch - vtxy;
  double zdist = trkvtxz_bestmatch - vtxz;
  if ( (xdist*xdist + ydist*ydist + zdist*zdist) < kRSphere) hasVertex = true;

  // is matched track vertex in our fiducial volume
  if (trkvtxx_bestmatch > kFidVolStartX && trkvtxx_bestmatch < kFidVolEndX
      && trkvtxy_bestmatch > kFidVolStartY && trkvtxy_bestmatch < kFidVolEndY
      && trkvtxz_bestmatch > kFidVolStartZ && trkvtxz_bestmatch < kFidVolEndZ) inFidVol = true;

  CrossingTrack(clusterlist, fmcls, isdata); 
  if (nearestz[2] > 4) noCrossingTrk = true;

  BrokenTrack(tracklist, isdata);
  if (largest_open_angle < 170) noBrokenTrk = true;
  else { // large opening angle and track end not contained
    if (trkendx_open_angle > kFidVolStartX && trkendx_open_angle < kFidVolEndX) 
      if (trkendy_open_angle > kFidVolStartY && trkendy_open_angle < kFidVolEndY) 
	if (trkendz_open_angle > kFidVolStartZ && trkendz_open_angle < kFidVolEndZ) 
	  noBrokenTrk = true;
  }

  // TESTING
  return  !SimpleMatch(minoslist, isdata);

  // for CC inclusive selection
  //  if (minosMatched && hasVertex && inFidVol && noCrossingTrk && noBrokenTrk) return true;
  //  else return false;

  // for inverse CC inclusive selection (make this a fcl parameter!!)
  if (minosMatched && hasVertex && inFidVol && noCrossingTrk && noBrokenTrk) return false;
  else return true;


} // filter

// ------------------------------------------------- 

void t962::ArgoneutCCInclusiveFilter::beginJob()
{
  // Implementation of optional member function here.
} // beginJob

// ------------------------------------------------- 

void t962::ArgoneutCCInclusiveFilter::endJob()
{
  // Implementation of optional member function here.
} // endJob

// ------------------------------------------------- 

void t962::ArgoneutCCInclusiveFilter::MatchMINOS(std::vector<art::Ptr<recob::Track>>& tracklist,
                                                 art::FindManyP<t962::MINOS>& fmmatch,
                                                 bool isdata) {
  double totaldiffmin = 1e10;

  for(size_t i=0; i<tracklist.size();++i){
    // No MINOS matched tracks found
    if (!fmmatch.at(i).size()) continue;
    
    TVector3 trackStart = tracklist[i]->Vertex<TVector3>();
    TVector3 trackEnd = tracklist[i]->End<TVector3>();
    TVector3 larStart = tracklist[i]->VertexDirection<TVector3>();
    TVector3 larEnd = tracklist[i]->EndDirection<TVector3>();
    
    //Redefine direction using track start/end point
    larEnd[0] = (trackEnd.X()-trackStart.X())/sqrt(pow(trackEnd.X()-trackStart.X(),2)+pow(trackEnd.Y()-trackStart.Y(),2)+pow(trackEnd.Z()-trackStart.Z(),2));
    larEnd[1] = (trackEnd.Y()-trackStart.Y())/sqrt(pow(trackEnd.X()-trackStart.X(),2)+pow(trackEnd.Y()-trackStart.Y(),2)+pow(trackEnd.Z()-trackStart.Z(),2));
    larEnd[2] = (trackEnd.Z()-trackStart.Z())/sqrt(pow(trackEnd.X()-trackStart.X(),2)+pow(trackEnd.Y()-trackStart.Y(),2)+pow(trackEnd.Z()-trackStart.Z(),2));
    
    if (trackStart.Z()>trackEnd.Z()){//flip direction
      auto tmp = trackStart;
      trackStart = trackEnd;
      trackEnd = tmp;
      for (size_t j = 0; j<3; ++j){
        double tmp = larStart[j];
        larStart[j] = -larEnd[j];
        larEnd[j] = -tmp;
      }
    }
    for (size_t j = 0; j<fmmatch.at(i).size(); ++j){
      
      double D=(90*0.5)+(42.4*2.54)-5.588; //distance from the front (upstream) of the TPC to the 1st Minos plane 
      //(this minus number is the one we measured with Mitch)
	
      double x_offset=117.4; // previously 116.9;
      double y_offset=19.3; // previously  20.28;
      
      if (isdata){
        x_offset += -3.134 + 1.867576;
        y_offset += -0.837;
      }
      else{
        x_offset += -0.893;
        y_offset += -1.597 + 2;
      }
      
      
      double dz = D - trackEnd.Z()+(100.0 * fmmatch.at(i)[j]->ftrkVtxZ);//z-difference between end of T962 track and
      //begin of MINOS track...in centimeters
      
      double l = dz/(larEnd[2]);//3-d distance between end of T962 track and begin of MINOS track
      
      double x_pred = l*larEnd[0]+trackEnd.X();//predicted x-pos. of T962 track at z-position equal to
      //start of MINOS track
      double y_pred = l*larEnd[1]+trackEnd.Y();//predicted y-pos. of T962 track at z-position equal to
      //start of MINOS track
      
      double dx = 100.0*fmmatch.at(i)[j]->ftrkVtxX - x_offset - x_pred;
      double dy = 100.0*fmmatch.at(i)[j]->ftrkVtxY + y_offset - y_pred;
      
      double rdiff = sqrt(dx*dx + dy*dy);
      double thetadiff = TMath::ACos((larEnd[0]*fmmatch.at(i)[j]->ftrkdcosx)+(larEnd[1]*fmmatch.at(i)[j]->ftrkdcosy)+(larEnd[2]*fmmatch.at(i)[j]->ftrkdcosz));
      //totaldiff is a measure of the agreement between the ArgoNeuT projected track and the MINOS track based on radial distance and angle. totaldiff= rdiff/cos(theta)  
      double totaldiff=fabs(rdiff/((larEnd[0]*fmmatch.at(i)[j]->ftrkdcosx)+(larEnd[1]*fmmatch.at(i)[j]->ftrkdcosy)+(larEnd[2]*fmmatch.at(i)[j]->ftrkdcosz)));
      if (totaldiff<totaldiffmin){
        totaldiffmin = totaldiff;
        bestmatch = i;
        bestmatchminos = j;
        rdiff_minos = rdiff;
        thetadiff_minos = thetadiff;
        xdiff_minos = dx;
        ydiff_minos = dy;
        //std::cout<<i<<" "<<rdiff<<" "<<thetadiff<<" "<<dx<<" "<<dy<<" "<<totaldiff<<" "<<bestmatch<<std::endl;
      }
    }//loop over all matched MINOS tracks

  }//loop over all argoneut tracks
  
} // MatchMINOS

// ------------------------------------------------- 

void t962::ArgoneutCCInclusiveFilter::VertexSelect(std::vector<art::Ptr<recob::Vertex>>& vertexlist,
						   std::vector<art::Ptr<recob::Track>>& tracklist,
						   bool isdata) {

  if (bestmatch < 0) return; // no minos match was found
  
  double mindist = 99999;

  // get matched track info
  recob::Track::Point_t trackStart, trackEnd;
  std::tie(trackStart, trackEnd) = tracklist[bestmatch]->Extent();

  trkvtxx_bestmatch = trackStart.X();
  trkvtxy_bestmatch = trackStart.Y();
  trkvtxz_bestmatch = trackStart.Z();

  trkendx_bestmatch = trackEnd.X();
  trkendy_bestmatch = trackEnd.Y();
  trkendz_bestmatch = trackEnd.Z();

  // flip track if need be
  if (trackStart.Z() > trackEnd.Z()) {
    trkvtxx_bestmatch = trackEnd.X();
    trkvtxy_bestmatch = trackEnd.Y();
    trkvtxz_bestmatch = trackEnd.Z();

    trkendx_bestmatch = trackStart.X();
    trkendy_bestmatch = trackStart.Y();
    trkendz_bestmatch = trackStart.Z();
  }

  for (size_t i = 0; i < vertexlist.size(); ++i) {
    double vtxxyz[3];
    vertexlist[i]->XYZ(vtxxyz);

    // is the vertex inside our fiducial volume?
    if (vtxxyz[0] < kFidVolStartX || vtxxyz[0] > kFidVolEndX) continue;
    if (vtxxyz[1] < kFidVolStartY || vtxxyz[1] > kFidVolEndY) continue;
    if (vtxxyz[2] < kFidVolStartZ || vtxxyz[2] > kFidVolEndZ) continue;

    double distx = vtxxyz[0] - trkvtxx_bestmatch;
    double disty = vtxxyz[1] - trkvtxy_bestmatch;
    double distz = vtxxyz[2] - trkvtxz_bestmatch;
    double dist = sqrt(distx*distx + disty*disty + distz*distz);
    
    if (dist < mindist) {
      mindist = dist;
      vertex_index = i;
      vtxx = vtxxyz[0];
      vtxy = vtxxyz[1];
      vtxz = vtxxyz[2];

    }
    
  } // loop through vertices

} // VertexSelect

// ------------------------------------------------- 

void t962::ArgoneutCCInclusiveFilter::CrossingTrack(std::vector<art::Ptr<recob::Cluster>>& clusterlist,
						    art::FindManyP<recob::Hit>& fmcls,
						    bool isdata) {

  art::ServiceHandle<geo::Geometry> geom;
  auto const* detprop = lar::providerFrom<detinfo::DetectorPropertiesService>();

  int nplanes = geom->Nplanes();
  std::vector< std::vector<int> > Cls(nplanes);
  for (size_t i = 0; i < clusterlist.size(); ++i) {
    switch (clusterlist[i]->View()) {
    case geo::kU :
      Cls[0].push_back(i);
      break;
    case geo::kV :
      Cls[1].push_back(i);
      break;
    default :
      break;
    } // end switch
  } // loop through clusters

  double mindistz = 99999;
  for (size_t i = 0; i < Cls[0].size(); ++i) {
    std::vector< art::Ptr<recob::Hit> > hitlistU = fmcls.at(Cls[0][i]);
    for (auto ihitU = hitlistU.begin(); ihitU != hitlistU.end(); ++ihitU) {
      double xU = detprop->ConvertTicksToX((*ihitU)->PeakTime(),0,0,0);
      for (size_t j = 0; j < Cls[1].size(); ++j) {
	std::vector< art::Ptr<recob::Hit> > hitlistV = fmcls.at(Cls[1][j]);
	for (auto ihitV = hitlistV.begin(); ihitV != hitlistV.end(); ++ihitV) {
	  double xV = detprop->ConvertTicksToX((*ihitV)->PeakTime(),1,0,0);
	  if (std::abs(xU-xV)<0.5) {
	    double y,z;
	    if (geom->ChannelsIntersect((*ihitU)->Channel(),(*ihitV)->Channel(), y, z)) {
	      if (std::abs(z) < mindistz) {
		mindistz = std::abs(z);
		nearestz[0] = xU;
		nearestz[1] = y;
		nearestz[2] = z;
	      }
	    }
	  }
	} // loop through hitV
      } // loop through Cls[1]
    } // loop through hitU
  } // loop through Cls[0]

} // CrossingTrack

// ------------------------------------------------- 

void t962::ArgoneutCCInclusiveFilter::BrokenTrack(std::vector<art::Ptr<recob::Track>>& tracklist,
                                                 bool isdata) {

  double lenx_trkminos = trkendx_bestmatch - trkvtxx_bestmatch;
  double leny_trkminos = trkendy_bestmatch - trkvtxy_bestmatch;
  double lenz_trkminos = trkendz_bestmatch - trkvtxz_bestmatch;
  double len_trkminos = sqrt(lenx_trkminos*lenx_trkminos + leny_trkminos*leny_trkminos + lenz_trkminos*lenz_trkminos);
  double trkminoscosx = lenx_trkminos/len_trkminos;
  double trkminoscosy = leny_trkminos/len_trkminos;
  double trkminoscosz = lenz_trkminos/len_trkminos;

  for (size_t i = 0; i < tracklist.size(); ++i) {
    if ((int)i == bestmatch) continue;
    
    recob::Track::Point_t trackStart, trackEnd;
    std::tie(trackStart, trackEnd) = tracklist[i]->Extent();

    double trkvtxx = trackStart.X();
    double trkvtxy = trackStart.Y();
    double trkvtxz = trackStart.Z();

    double trkendx = trackEnd.X();
    double trkendy = trackEnd.Y();
    double trkendz = trackEnd.Z();

    double dist1x = trkvtxx_bestmatch - trkvtxx;
    double dist1y = trkvtxy_bestmatch - trkvtxy;
    double dist1z = trkvtxz_bestmatch - trkvtxz;

    double dist2x = trkvtxx_bestmatch - trkendx;
    double dist2y = trkvtxy_bestmatch - trkendy;
    double dist2z = trkvtxz_bestmatch - trkendz;

    double dist1 = sqrt(dist1x*dist1x + dist1y*dist1y + dist1z*dist1z);
    double dist2 = sqrt(dist2x*dist2x + dist2y*dist2y + dist2z*dist2z);

    // flip track if need be
    if (dist1 > dist2) {
      double tempx = trkvtxx; double tempy = trkvtxy; double tempz = trkvtxz;
      trkvtxx = trkendx; trkvtxy = trkendy; trkvtxz = trkendz;
      trkendx = tempx; trkendy = tempy; trkendz = tempz;
    } // flip track

    double x = trkvtxx_bestmatch - trkvtxx;
    double y = trkvtxy_bestmatch - trkvtxy;
    double z = trkvtxz_bestmatch - trkvtxz;
    if (x*x + y*y + z*z > kRSphere) continue;

    // calculate angle of track
    double lenx = trkendx - trkvtxx;
    double leny = trkendy - trkvtxy;
    double lenz = trkendz - trkvtxz;
    double len = std::sqrt(lenx*lenx + leny*leny + lenz*lenz);

    double trkvtxcosx = lenx/len;
    double trkvtxcosy = leny/len;
    double trkvtxcosz = lenz/len;

    double dotprod = trkvtxcosx*trkminoscosx + trkvtxcosy*trkminoscosy + trkvtxcosz*trkminoscosz;
    double arccos = std::acos(dotprod) * 180.0/3.14159265358979; // convert to degrees

    if (arccos > 180) arccos = 360 - arccos;
    if (arccos < 0) arccos = -arccos;

    if (arccos < largest_open_angle) continue;

    largest_open_angle = arccos;
    trkendx_open_angle = trkendx;
    trkendy_open_angle = trkendy;
    trkendz_open_angle = trkendz;

  } // loop through tracks

} // BrokenTrack

// ------------------------------------------------- 

bool t962::ArgoneutCCInclusiveFilter::SimpleMatch(std::vector<art::Ptr<t962::MINOS>>& minoslist, bool isdata) {
  
  double D=(90*0.5)+(42.4*2.54)-5.588; //distance from the front (upstream) of the TPC to the 1st Minos plane         
  //(this minus number is the one we measured with Mitch)                                                             
  double x_offset=117.4; // previously 116.9;                                                                         
  double y_offset=19.3; // previously  20.28;                                                                         

  if (isdata){
    x_offset += -3.134 + 1.867576;
    y_offset += -0.837;
  }
  else{
    x_offset += -0.893;
    y_offset += -1.597 + 2;
  }

  for (size_t i = 0; i < minoslist.size(); ++i) {
    for (int j = 0; j < 10; ++j) {
      
      double zInArgoNeuT = (double)j * 10;
      
      // calculate locatoin in ArgoNeuT
      double dz = - (D - zInArgoNeuT +(100.0 * minoslist[i]->ftrkVtxZ) ); // just the opposite of the function above
      double l = dz / minoslist[i]->ftrkdcosz;

      //predicted x-pos. of MINOS track at 10*j cm into ArgoNeuT        
      double x_pred = l*minoslist[i]->ftrkdcosx + 100.0*minoslist[i]->ftrkVtxX;
      //predicted y-pos. of MINOS track at 10*j cm into ArgoNeuT
      double y_pred = l*minoslist[i]->ftrkdcosy + 100.0*minoslist[i]->ftrkVtxY;
      
      x_pred = x_pred - x_offset;
      y_pred = y_pred + y_offset;
      //      std::cout << "(x, y) = (" << x_pred << "," << y_pred << ")" << std::endl;

      if (x_pred > kFidVolStartX && x_pred < kFidVolEndX)
	if (y_pred > kFidVolStartY && y_pred < kFidVolEndY)
	  return true;
    } // iterate through z
  } // iterate through MINOS tracks

  return false;

} // SimpleMatch

// ------------------------------------------------- 

DEFINE_ART_MODULE(t962::ArgoneutCCInclusiveFilter)
